#include "ui.h"
#include "gpu.h"
#include "draw.h"

#include "arpa/inet.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"

#include "ft2build.h"
#include FT_FREETYPE_H

VkResult test_ui(GLFWwindow* window, UIContext* ui, RenderContext* render) {
  uint32_t font_index;
  uint32_t background_index;
  FT_Library library;
  VkResult result;

  if(FT_Init_FreeType(&library) != FT_Err_Ok) {
    return VK_ERROR_UNKNOWN;
  }
  VK_RESULT(load_font("test.ttf", 16, VK_TRUE, library, render, ui, &font_index));

  GPUString context_strings[] = {
    {
      .pos = {0, 16},
      .size = 16,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 0,
      .length = strlen("Example"),
      .font = font_index,
    },
  };

  uint32_t context_codes[100];
  map_string("Example Text", context_codes, 0, ui->fonts[0].charmap, ui->fonts[0].num_symbols);

  LayerInput context_layers[] = {
    {
      .num_strings = 1,
      .strings = context_strings,
      .num_codes = 100,
      .codes = context_codes,
    },
  };

  ContainerInput context_info = {
    .id = 0x01,
    .anchor = ANCHOR_TOP_LEFT,
    .size = {WINDOW_MIN_WIDTH, WINDOW_MIN_HEIGHT},
    .layer_count = sizeof(context_layers)/sizeof(LayerInput),
    .layers = context_layers,
  };

  vec2 map_size = {250, 200};

  GPUDrawable map_rects[] = {
    {
      .pos = {0.0, 0.0},
      .size = {map_size[0], map_size[1]},
      .color = {0.0, 0.5,0.8, 1.0},
      .type = 0,
    },
  };

  LayerInput map_layers[] = {
    {
      .num_drawables = sizeof(map_rects)/sizeof(GPUDrawable),
      .drawables = map_rects,
    },
  };

  ContainerInput map_info = {
    .id = 0x02,
    .anchor = ANCHOR_TOP_RIGHT,
    .size = {map_size[0], map_size[1]},
    .layer_count = sizeof(map_layers)/sizeof(LayerInput),
    .layers = map_layers,
  };

  vec2 inventory_size = {map_size[0], WINDOW_MIN_HEIGHT-map_size[1]};

  GPUDrawable inventory_rects[] = {
    {
      .pos = {0, 0},
      .size = {inventory_size[0], inventory_size[1]},
      .color = {0.5, 0.8, 0.0, 1.0},
      .type = 0,
    },
  };

  LayerInput inventory_layers[] = {
    {
      .num_drawables = sizeof(inventory_rects)/sizeof(GPUDrawable),
      .drawables = inventory_rects,
    },
  };

  ContainerInput inventory_info = {
    .id = 0x03,
    .anchor = ANCHOR_BOTTOM_RIGHT,
    .size = {inventory_size[0], inventory_size[1]},
    .layer_count = sizeof(inventory_layers)/sizeof(LayerInput),
    .layers = inventory_layers,
  };

  vec2 chat_size = {WINDOW_MIN_WIDTH-inventory_size[0], 200};

  GPUDrawable chat_rects[] = {
    {
      .pos = {0, 0},
      .size = {chat_size[0], chat_size[1]},
      .color = {0.8, 0.0, 0.5, 1.0},
      .type = 0,
    },
  };

  LayerInput chat_layers[] = {
    {
      .num_drawables = sizeof(chat_rects)/sizeof(GPUDrawable),
      .drawables = chat_rects,
    },
  };

  ContainerInput chat_info = {
    .id = 0x04,
    .anchor = ANCHOR_BOTTOM_LEFT,
    .size = {chat_size[0], chat_size[1]},
    .layer_count = sizeof(chat_layers)/sizeof(LayerInput),
    .layers = chat_layers,
  };

  VK_RESULT(create_container(&context_info, render, ui));
  VK_RESULT(create_container(&map_info, render, ui));
  VK_RESULT(create_container(&inventory_info, render, ui));
  VK_RESULT(create_container(&chat_info, render, ui));

  VK_RESULT(record_draw_commands(render, ui));

  return VK_SUCCESS;
}

VkResult render_thread(GLFWwindow* window, RenderContext* render, UIContext* ui) {

  VkResult result;

  double last_frame_time = glfwGetTime();
  while(glfwWindowShouldClose(window) == 0) {
    glfwPollEvents();

    double frame_time = glfwGetTime();
    double delta_time = frame_time - last_frame_time;
    result = draw_frame(render, frame_time);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      return result;
    }
    last_frame_time = frame_time;
  }
  return 0;
}

typedef enum ClientAuthStateEnum {
  AUTH_SAVED,
  AUTH_CREDENTIALS,
  AUTH_ATTEMPT,
  AUTH_ERROR,
  AUTH_SUCCESS,
} ClientAuthState;

typedef struct ClientStateStruct {
  
} ClientState;

int logic_thread() {
  return 0;
}

int network_thread() {
  return 0;
}

void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
  (void)window;
  (void)key;
  (void)scancode;
  (void)action;
  (void)mods;
}

void mouse_button_callback(GLFWwindow* window, int button, int action, int mods) {
  (void)mods;
  switch(button) {
  // Handle camera hover
  case GLFW_MOUSE_BUTTON_MIDDLE:
    if(action == GLFW_PRESS) {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
    } else {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }
    break;
  case GLFW_MOUSE_BUTTON_LEFT:
    if(action == GLFW_PRESS) {
    }
    break;
  case GLFW_MOUSE_BUTTON_RIGHT:
    if(action == GLFW_PRESS) {
    }
    break;
  }
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset) {
  (void)window;
  (void)xoffset;
  (void)yoffset;
}

void cursor_pos_callback(GLFWwindow* window, double xpos, double ypos) {
  (void)window;
  (void)xpos;
  (void)ypos;
}

int main() {
  GLFWwindow* window = init_window();
  if(window == NULL) {
    return 1;
  }

  glfwSetKeyCallback(window, key_callback);
  glfwSetMouseButtonCallback(window, mouse_button_callback);
  glfwSetScrollCallback(window, scroll_callback);
  glfwSetCursorPosCallback(window, cursor_pos_callback);

  RenderContext render = {};
  UIContext ui = {};

  VkResult result;
  VK_RESULT(init_vulkan(window, &render));
  
  // TODO: make # of fonts/textures/containers scaling, recreate GPU buffers as necessary
  VK_RESULT(create_ui_context(10, 10, 10, &render, &ui));

  //////////////////////////////////
  /// Test Code
  //////////////////////////////////

  VK_RESULT(test_ui(window, &ui, &render));

  //////////////////////////////////

  if(render_thread(window, &render, &ui) != VK_SUCCESS) {
    return 3;
  }

  return 0;
}
